# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading


def setup_logger():
    logger = logging.getLogger("sapientml")
    mutex = threading.Lock()
    if len(logger.handlers) == 0:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("[%(asctime)s] %(levelname)s:%(message)s", "%Y-%m-%d %H:%M:%S"))
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
    mutex.acquire()
    return logger



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math
from datetime import date, datetime, timedelta
from uuid import UUID

import numpy as np


class JSONEncoder(json.JSONEncoder):
    """Encoding Json"""

    def default(self, obj):
        """default method for JSONEncoder.

        Parameters
        ----------
        obj : Instance of the object

        Results
        ------
        Depends upon the instance type
            if instance is integer, it will return int.
            if instance is float, it will return float or None if it's a NaN value.
            if instance is string, it will return string.
            if instance is UUID, it will returns string.

        """
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, timedelta):
            return str(obj)
        if isinstance(obj, Exception):
            return repr(obj)
        return super(JSONEncoder, self).default(obj)

    def encode(self, obj):
        """encode method for JSONEncoder.

        Parameters
        ----------
        obj : The object to be encoded as a JSON string

        Results
        ------
        Replaces any float 'nan' values with 'None' in the input object.

        """

        def replace_nan(o):
            if isinstance(o, float) and math.isnan(o):
                return None
            if isinstance(o, list):
                return [replace_nan(item) for item in o]
            if isinstance(o, dict):
                return {key: replace_nan(value) for key, value in o.items()}
            return o

        obj = replace_nan(obj)
        return super().encode(obj)


class JSONDecoder(json.JSONDecoder):
    """Decoding Json"""

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        """object_hook for JSONDecoder.

        Parameters
        ----------
        obj : Instance of the object

        Results
        ------
        obj conducted the following replacement:
        - "nan" with np.nan
        - "inf" with np.inf
        - "-inf" with -np.inf

        """

        def replace(obj, original, replaced):
            if isinstance(obj, dict):
                return {k: replace(v, original, replaced) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [replace(v, original, replaced) for v in obj]
            elif obj == original:
                return replaced
            return obj

        obj = replace(obj, "nan", np.nan)
        obj = replace(obj, "inf", np.inf)
        obj = replace(obj, "-inf", -np.inf)

        return obj



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from os import PathLike
from pathlib import Path
from typing import Literal, Optional, Union

import pandas as pd

from ..executor import run
from ..params import save_file
from ..util.logging import setup_logger

logger = setup_logger()


class GeneratedModel:
    def __init__(
        self,
        input_dir: PathLike,
        save_datasets_format: Literal["csv", "pickle"],
        timeout: int,
        csv_encoding: str,
        csv_delimiter: str,
        params: dict,
    ):
        """
        The constructor of GeneratedModel.
        Instantiating this class by yourself is not intended.

        Parameters
        ----------
        input_dir: PathLike
            Directory path containing training/prediction scripts and trained models.
        save_datasets_format: 'csv' or 'pickle'
            Data format when the input dataframes are written to files.
            Ignored when all inputs are specified as file path.
        timeout: int
            Timeout for the execution of training and prediction.
        csv_encoding: str
            Encoding method when csv files are involved.
            Ignored when only pickle files are involved.
        csv_delimiter: str
            Delimiter to read csv files.
        params: dict
            All the parameters of SapientML and plugins
        """

        self.files = dict()
        self.save_datasets_format = save_datasets_format
        self.timeout = timeout
        self.csv_encoding = csv_encoding
        self.csv_delimiter = csv_delimiter
        self.params = params
        input_dir = Path(input_dir)
        self._readfile(input_dir / "final_script.py", input_dir)
        self._readfile(input_dir / "final_train.py", input_dir)
        self._readfile(input_dir / "final_predict.py", input_dir)

        for filepath in input_dir.glob("lib/*.py"):
            self._readfile(filepath, input_dir)

        for filepath in input_dir.glob("**/*.pkl"):
            if save_datasets_format == "pickle" and "training.pkl" == filepath.name:
                continue
            self._readfile(filepath, input_dir)

    @staticmethod
    def create(
        input_dir: PathLike,
        save_datasets_format: Literal["csv", "pickle"],
        timeout: int,
        csv_encoding: str,
        csv_delimiter: str,
        params: dict,
    ):
        """
        The factory method of GeneratedModel.
        Instantiating this class by yourself is not intended.

        Parameters
        ----------
        input_dir: PathLike
            Directory path containing training/prediction scripts and trained models.
        save_datasets_format: 'csv' or 'pickle'
            Data format when the input dataframes are written to files.
            Ignored when all inputs are specified as file path.
        timeout: int
            Timeout for the execution of training and prediction.
        csv_encoding: str
            Encoding method when csv files are involved.
            Ignored when only pickle files are involved.
        csv_delimiter: str
            Delimiter to read csv files.
        params: dict
            All the parameters of SapientML and plugins
        """
        return GeneratedModel(input_dir, save_datasets_format, timeout, csv_encoding, csv_delimiter, params)

    def _readfile(self, filepath, input_dir):
        with open(filepath, "rb") as f:
            self.files[str(filepath.relative_to(input_dir))] = f.read()

    def save(self, output_dir: PathLike):
        """
        Save generated code to `output_dir` folder

        Parameters
        ----------
        output_dir: Path-like object
            Training dataframe.

        Returns
        -------
        self: GeneratedModel
            GeneratedModel object itself
        """
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        for filename, content in self.files.items():
            Path(output_dir / filename).parent.mkdir(exist_ok=True, parents=True)
            with open(output_dir / filename, "wb") as f:
                f.write(content)

    @staticmethod
    def load(id: str):
        pass

    def fit(self, X: pd.DataFrame, y: Optional[Union[pd.DataFrame, pd.Series]] = None):
        """
        Generate ML scripts for input data.

        Parameters
        ----------
        X: pandas.DataFrame
            Training dataframe. Contains target values if `y` is `None`.
        y: pandas.DataFrame or pandas.Series
            The target values.

        Returns
        -------
        self: GeneratedModel
            GeneratedModel object itself
        """
        if y is not None:
            X = pd.concat([X, y], axis=1)

        with tempfile.TemporaryDirectory() as temp_dir_path_str:
            temp_dir = Path(temp_dir_path_str).absolute()
            temp_dir.mkdir(exist_ok=True)
            self.save(temp_dir)
            filename = "training." + ("pkl" if self.save_datasets_format == "pickle" else "csv")
            save_file(X, str(temp_dir / filename), self.csv_encoding, self.csv_delimiter)
            logger.info("Building model by generated pipeline...")
            result = run(str(temp_dir / "final_train.py"), self.timeout)
            if result.returncode != 0:
                raise RuntimeError(f"Training was failed due to the following Error: {result.error}")
            for filepath in temp_dir.glob("**/*.pkl"):
                if self.save_datasets_format == "pickle" and "training.pkl" == filepath.name:
                    continue
                self._readfile(filepath, temp_dir)
        return self

    def predict(self, X: pd.DataFrame):
        """Predicts the output of the test_data and store in the prediction_result.csv.

        Parameters
        ---------
        X: pd.DataFrame
            Dataframe used for predicting the result.

        Returns
        -------
        result_df : pd.DataFrame
            It returns the prediction_result.csv result in dataframe format.
        """
        with tempfile.TemporaryDirectory() as temp_dir_path_str:
            temp_dir = Path(temp_dir_path_str).absolute()
            temp_dir.mkdir(exist_ok=True)
            self.save(temp_dir)
            filename = "test." + ("pkl" if self.save_datasets_format == "pickle" else "csv")
            save_file(X, str(temp_dir / filename), self.csv_encoding, self.csv_delimiter)
            result = run(str(temp_dir / "final_predict.py"), self.timeout)
            if result.returncode != 0:
                raise RuntimeError(f"Prediction was failed due to the following Error: {result.error}")
            id_columns_for_prediction = self.params.get("id_columns_for_prediction", [])
            if not id_columns_for_prediction:
                id_columns_for_prediction = [0]
            result_df = pd.read_csv(temp_dir / "prediction_result.csv", index_col=id_columns_for_prediction)
            return result_df



# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import tempfile
import time
from os import PathLike
from pathlib import Path
from typing import Literal, Optional, Union
from uuid import uuid4 as uuid

import lancedb
import pandas as pd
from lancedb.pydantic import LanceModel
from pydantic import computed_field

from ..executor import run
from ..params import save_file
from ..util.logging import setup_logger

logger = setup_logger()


def _readfile(files, filepath: Path, input_dir: Path, encoding: str):
    _, ext = os.path.splitext(filepath)
    if ext == ".pkl":
        with open(filepath, "rb") as f:
            files.append((str(filepath.relative_to(input_dir)), " ".join(map(str, f.read()))))
    else:
        with open(filepath, "r", encoding=encoding) as f:
            files.append((str(filepath.relative_to(input_dir)), f.read()))


class GeneratedModel(LanceModel):
    id: str
    prev_id: str | None = None
    timestamp: float
    files: list[tuple[str, str]]
    save_datasets_format: str
    timeout: int
    csv_encoding: str
    csv_delimiter: str
    params_list: list[tuple[str, str, str]]
    score: float
    metric: str

    @staticmethod
    def create(
        input_dir: PathLike,
        save_datasets_format: Literal["csv", "pickle"],
        timeout: int,
        csv_encoding: str,
        csv_delimiter: str,
        params: dict,
    ):
        """
        The factory method of GeneratedModel.
        Instantiating this class by yourself is not intended.

        Parameters
        ----------
        input_dir: PathLike
            Directory path containing training/prediction scripts and trained models.
        save_datasets_format: 'csv' or 'pickle'
            Data format when the input dataframes are written to files.
            Ignored when all inputs are specified as file path.
        timeout: int
            Timeout for the execution of training and prediction.
        csv_encoding: str
            Encoding method when csv files are involved.
            Ignored when only pickle files are involved.
        csv_delimiter: str
            Delimiter to read csv files.
        params: dict
            All the parameters of SapientML and plugins
        """
        input_dir = Path(input_dir)
        files = list()
        _readfile(files, input_dir / "final_script.py", input_dir, csv_encoding)
        _readfile(files, input_dir / "final_train.py", input_dir, csv_encoding)
        _readfile(files, input_dir / "final_predict.py", input_dir, csv_encoding)

        for filepath in input_dir.glob("lib/*.py"):
            _readfile(files, filepath, input_dir, csv_encoding)

        for filepath in input_dir.glob("**/*.pkl"):
            if save_datasets_format == "pickle" and "training.pkl" == filepath.name:
                continue
            _readfile(files, filepath, input_dir, csv_encoding)

        with open(input_dir / "final_script.out.json", "r") as f:
            data = json.load(f)
            score = float(data["score"])
            metric = data["metric"]

        id = str(uuid())

        return GeneratedModel(
            id=id,
            timestamp=time.time(),
            files=files,
            save_datasets_format=save_datasets_format,
            timeout=timeout,
            csv_encoding=csv_encoding,
            csv_delimiter=csv_delimiter,
            params_list=[(k, str(type(v)), str(v)) for k, v in params.items()],
            score=score,
            metric=metric,
        )

    @computed_field(repr=False)
    @property
    def params(self) -> dict:
        return {k: v if "str" in t else eval(v) for k, t, v in self.params_list}

    def save(self, output_dir: PathLike = ".lancedb"):
        """
        Save itself to database
        """
        db = lancedb.connect(output_dir)
        table = db.create_table("artifacts", schema=GeneratedModel, exist_ok=True)
        self.prev_id = self.id
        self.id = str(uuid())
        self.timestamp = time.time()
        table.add([self])
        return self.id

    @staticmethod
    def load(id: str):
        """
        Load a GeneratedModel with id from database

        Parameters
        ----------
        id: str
            UUID string of the model to be loaded

        Returns
        -------
        model: GeneratedModel
            GeneratedModel object loaded
        """
        db = lancedb.connect(".lancedb")
        table = db.open_table("artifacts")
        model = table.search().where(f"id='{id}'").limit(1).to_pydantic(GeneratedModel)[0]
        return model

    def export(self, output_dir: Union[str, PathLike]):
        """
        Export generated code to `output_dir` folder

        Parameters
        ----------
        output_dir: Path-like object
            Training dataframe.

        Returns
        -------
        self: GeneratedModel
            GeneratedModel object itself
        """
        output_dir = Path(output_dir)
        if output_dir.exists():
            shutil.rmtree(output_dir)
        output_dir.mkdir(parents=True)
        (output_dir / "lib").mkdir(exist_ok=True)
        for filename, content in self.files:
            _, ext = os.path.splitext(filename)
            if ext == ".pkl":
                with open(output_dir / filename, "wb") as f:
                    f.write(b"".join([int(i).to_bytes(1, "little") for i in content.split(" ")]))
            else:
                with open(output_dir / filename, "w", encoding=self.csv_encoding) as f:
                    f.write(content)

    def fit(self, X: pd.DataFrame, y: Optional[Union[pd.DataFrame, pd.Series]] = None):
        """
        Generate ML scripts for input data.

        Parameters
        ----------
        X: pandas.DataFrame
            Training dataframe. Contains target values if `y` is `None`.
        y: pandas.DataFrame or pandas.Series
            The target values.

        Returns
        -------
        self: GeneratedModel
            GeneratedModel object itself
        """
        if y is not None:
            X = pd.concat([X, y], axis=1)

        with tempfile.TemporaryDirectory() as temp_dir_path_str:
            temp_dir = Path(temp_dir_path_str).absolute()
            temp_dir.mkdir(exist_ok=True)
            self.export(temp_dir)
            filename = "training." + ("pkl" if self.save_datasets_format == "pickle" else "csv")
            save_file(X, str(temp_dir / filename), self.csv_encoding, self.csv_delimiter)
            logger.info("Building model by generated pipeline...")
            result = run(str(temp_dir / "final_train.py"), self.timeout)
            if result.returncode != 0:
                raise RuntimeError(f"Training was failed due to the following Error: {result.error}")
            for filepath in temp_dir.glob("**/*.pkl"):
                if self.save_datasets_format == "pickle" and "training.pkl" == filepath.name:
                    continue
                _readfile(self.files, filepath, temp_dir, self.csv_encoding)
        return self

    def predict(self, X: pd.DataFrame):
        """Predicts the output of the test_data and store in the prediction_result.csv.

        Parameters
        ---------
        X: pd.DataFrame
            Dataframe used for predicting the result.

        Returns
        -------
        result_df : pd.DataFrame
            It returns the prediction_result.csv result in dataframe format.
        """
        with tempfile.TemporaryDirectory() as temp_dir_path_str:
            temp_dir = Path(temp_dir_path_str).absolute()
            temp_dir.mkdir(exist_ok=True)
            self.export(temp_dir)
            filename = "test." + ("pkl" if self.save_datasets_format == "pickle" else "csv")
            save_file(X, str(temp_dir / filename), self.csv_encoding, self.csv_delimiter)
            result = run(str(temp_dir / "final_predict.py"), self.timeout)
            if result.returncode != 0:
                raise RuntimeError(f"Prediction was failed due to the following Error: {result.error}")
            id_columns_for_prediction = None
            for k, _, v in self.params_list:
                if k == "id_columns_for_prediction":
                    id_columns_for_prediction = eval(v)
                    break
            if id_columns_for_prediction is None:
                id_columns_for_prediction = [0]
            result_df = pd.read_csv(temp_dir / "prediction_result.csv", index_col=id_columns_for_prediction)
            return result_df



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import re

TASK_CLASSIFICATION = "classification"
TASK_REGRESSION = "regression"
PRED_DEFAULT = "default"  # predict()
PRED_PROBABILITY = "probability"  # predict_proba() for classifications


class Metric(enum.Enum):
    """metrics used for verifying the results accuracy.

    Attributes
    ----------
    F1 : str
        F1 Score metrics
    R2 : str
        R2 score used to evaluate the performance of a regression-based machine learning model.
    RMSLE : str
        Root Mean Squared Logarithmic Error (RMSLE)
    RMSE : str
        Root Mean Squared Error (RMSE)
    AUC : str
        AUC (Area Under the Curve)
    Accuracy : str
        Accuracy is a metric that generally describes how the model performs across all classes
        It is calculated as the ratio between the number of correct predictions to the total number of predictions.
    MAE : str
        Mean Absolute Error (MAE)
    MAPE : str
        Mean Absolute Percentage Error (MAPE)
    Gini : str
        The Gini Coefficient is used to evaluate the performance of Binary Classifier Models.
        The value of the Gini Coefficient can be between 0 to 1. The higher the Gini coefficient, the better is the model.
    LogLoss : str
        Log loss, also known as logarithmic loss or cross-entropy loss, is a common evaluation metric for binary classification models.
    ROC_AUC : str
        ROC (receiver operating characteristic curve)
        AUC (Area Under the Curve)
        ROC is a probability curve and AUC represents the degree or measure of separability.
    MCC : str
        The Matthews correlation coefficient (MCC)
    MAP_K : str
        Mean Average Precision at K (MAP@K) is one of the most commonly used evaluation metrics for recommender systems and other ranking related classification tasks.
    QWK : str
        QWK (Quadratic weighted kappa) is useful when dealing with imbalanced datasets or when the cost of misclassification varies across classes.

    """

    F1 = "f1"
    R2 = "r2"
    RMSLE = "RMSLE"
    RMSE = "RMSE"
    AUC = "auc"
    Accuracy = "accuracy"
    MAE = "MAE"
    Gini = "Gini"
    LogLoss = "LogLoss"
    ROC_AUC = "ROC_AUC"
    MCC = "MCC"  # Matthews correlation coefficient
    MAP_K = "MAP_"
    QWK = "QWK"  # Quadratic weighted kappa
    MAPE = "MAPE"

    @staticmethod
    def get(string) -> str:
        """get method.

        Parameters
        ----------
        string : str

        Results
        -------
        str
            It returns string.

        """
        ret = [x.value for x in Metric if string.lower() == x.name.lower()]
        if len(ret) == 1:
            return ret[0]
        elif re.match(rf"{Metric.MAP_K.value}[0-9]+", string):
            return string
        raise NotImplementedError(f"metric name '{string}' is not defined. Available: {str([x.name for x in Metric])}")

    @staticmethod
    def get_default_value(task_type: str) -> str:
        """get_default_value method.

        Parameters
        ----------
        task_type : str

        Results
        -------
        str
            It returns Metric.F1.value.

        """
        # NOTE: See pipeline_template.Pipeline.create_evaluation_code defines the default value of adaptation_metric.
        if task_type == TASK_REGRESSION:
            return Metric.R2.value
        # if task_type == TASK_CLASSIFICATION:
        return Metric.F1.value

    @staticmethod
    def metric_match_task_type(adaptation_metric: str, task_type: str) -> bool:
        """metric_match_task_type method.

        Parameters
        ----------
        adaptation_metric : str
        task_type : str

        Results
        -------
        bool
            True and otherwise False.

        """
        if task_type == "regression":
            if adaptation_metric in metrics_for_classification:
                return False
            elif re.match(rf"{Metric.MAP_K.value}[0-9]+", adaptation_metric):
                return False
        elif task_type == "classification" and adaptation_metric in metrics_for_regression:
            return False
        return True

    @staticmethod
    def metric_support_multioutput(adaptation_metric: str) -> bool:
        """metric_support_multioutput method.

        Parameters
        ----------
        adaptation_metric : str

        Results
        -------
        bool
            True and otherwise False.

        """
        if adaptation_metric in metrics_not_support_multioutput:
            return False
        elif re.match(rf"{Metric.MAP_K.value}[0-9]+", adaptation_metric):
            return False
        else:
            return True

    @staticmethod
    def metric_support_multiclass_multioutput(adaptation_metric: str) -> bool:
        """metric_support_multiclass_multioutput method.

        Parameters
        ----------
        adaptation_metric : str

        Results
        -------
        bool
            True and otherwise False.

        """
        if adaptation_metric is Metric.Gini.value:
            return True
        else:
            return False


metric_lower_is_better = [
    Metric.RMSLE.value,
    Metric.RMSE.value,
    Metric.MAE.value,
    Metric.LogLoss.value,
    Metric.MAPE.value,
]

metric_needing_predict_proba = [
    Metric.LogLoss.value,
    Metric.AUC.value,
    Metric.ROC_AUC.value,
    Metric.MAP_K.value,
    Metric.Gini.value,
]

metrics_for_regression = [Metric.R2.value, Metric.RMSLE.value, Metric.RMSE.value, Metric.MAE.value, Metric.MAPE.value]

metrics_for_classification = [
    Metric.F1.value,
    Metric.AUC.value,
    Metric.Accuracy.value,
    Metric.Gini.value,
    Metric.LogLoss.value,
    Metric.ROC_AUC.value,
    Metric.MCC.value,
    Metric.MAP_K.value,
    Metric.QWK.value,
]

metrics_not_support_multioutput = [
    Metric.MCC.value,
    Metric.MAP_K.value,
    Metric.QWK.value,
]
